#include "integral_channels_detector.cu.hpp"

#include "helpers/gpu/cuda_safe_call.hpp"

#include "cudatemplates/array.hpp"
#include "cudatemplates/symbol.hpp"
#include "cudatemplates/copy.hpp"

#include <cudatemplates/hostmemoryheap.hpp>

#include <boost/cstdint.hpp>

#include <stdexcept>



namespace {

/// small helper function that computes "static_cast<int>(ceil(static_cast<float>(total)/grain))", but faster
static inline int div_up(const int total, const int grain)
{
    return (total + grain - 1) / grain;
}


} // end of anonymous namespace


namespace doppia {
namespace objects_detection {

using namespace cv::gpu;

typedef Cuda::DeviceMemory<gpu_integral_channels_t::Type, 1>::Texture gpu_integral_channels_1d_texture_t;
gpu_integral_channels_1d_texture_t integral_channels_1d_texture;

typedef Cuda::DeviceMemory<gpu_integral_channels_t::Type, 2>::Texture gpu_integral_channels_2d_texture_t;
gpu_integral_channels_2d_texture_t integral_channels_2d_texture;

// global variable to switch from using 1d textures to using 2d textures
// On visics-gt680r 1d texture runs at ~4.8 Hz; 2d texture runs at ~4.4 Hz
//const bool use_2d_texture = false;
const bool use_2d_texture = true;

/// this method will do zero memory checks, the user is responsible of avoiding out of memory accesses
inline
__device__
float get_feature_value_global_memory(const IntegralChannelsFeature &feature,
                                      const int x, const int y,
                                      const gpu_integral_channels_t::KernelConstData &integral_channels)
{
    const IntegralChannelsFeature::rectangle_t &box = feature.box;

    const size_t
            &channel_stride = integral_channels.stride[1],
            &row_stride = integral_channels.stride[0];

    // if x or y are too high, some of these indices may be fall outside the channel memory
    const size_t
            channel_offset = feature.channel_index*channel_stride,
            top_left_index     = (x + box.min_corner().x()) + ((y + box.min_corner().y())*row_stride) + channel_offset,
            top_right_index    = (x + box.max_corner().x()) + ((y + box.min_corner().y())*row_stride) + channel_offset,
            bottom_left_index  = (x + box.min_corner().x()) + ((y + box.max_corner().y())*row_stride) + channel_offset,
            bottom_right_index = (x + box.max_corner().x()) + ((y + box.max_corner().y())*row_stride) + channel_offset;

    const gpu_integral_channels_t::Type
            a = integral_channels.data[top_left_index],
            b = integral_channels.data[top_right_index],
            c = integral_channels.data[bottom_right_index],
            d = integral_channels.data[bottom_left_index];

    const float feature_value = a +c -b -d;

    return feature_value;
}


inline
__device__
float get_feature_value_tex1d(const IntegralChannelsFeature &feature,
                              const int x, const int y,
                              const gpu_integral_channels_t::KernelConstData &integral_channels)
{
    const IntegralChannelsFeature::rectangle_t &box = feature.box;

    const size_t
            &channel_stride = integral_channels.stride[1],
            &row_stride = integral_channels.stride[0];

    // if x or y are too high, some of these indices may be fall outside the channel memory
    const size_t
            channel_offset = feature.channel_index*channel_stride;
    const size_t
            top_left_index     = (x + box.min_corner().x()) + ((y + box.min_corner().y())*row_stride) + channel_offset,
            top_right_index    = (x + box.max_corner().x()) + ((y + box.min_corner().y())*row_stride) + channel_offset,
            bottom_left_index  = (x + box.min_corner().x()) + ((y + box.max_corner().y())*row_stride) + channel_offset,
            bottom_right_index = (x + box.max_corner().x()) + ((y + box.max_corner().y())*row_stride) + channel_offset;

    // in CUDA 5 (4.2 ?) references to textures are not allowed, we use macro work around
    //    gpu_integral_channels_1d_texture_t &tex = integral_channels_1d_texture;
#define tex integral_channels_1d_texture
    //const gpu_integral_channels_t::Type  // could cause overflows during a + c
    // tex1Dfetch should be used to access linear memory (not text1D)
    const float
            a = tex1Dfetch(tex, top_left_index),
            b = tex1Dfetch(tex, top_right_index),
            c = tex1Dfetch(tex, bottom_right_index),
            d = tex1Dfetch(tex, bottom_left_index);
#undef tex


    const float feature_value = a +c -b -d;

    return feature_value;
}




template <typename FeatureType>
inline
__device__
float get_feature_value_tex2d(const FeatureType &feature,
                              const int x, const int y,
                              const gpu_3d_integral_channels_t::KernelConstData &integral_channels)
{
    // if x or y are too high, some of these indices may be fall outside the channel memory

    const size_t integral_channels_height = integral_channels.size[1];
    const float y_offset = y + feature.channel_index*integral_channels_height;

    // in CUDA 5 (4.2 ?) references to textures are not allowed, we use macro work around
    //    gpu_integral_channels_2d_texture_t &tex = integral_channels_2d_texture;
#define tex integral_channels_2d_texture

    const typename FeatureType::rectangle_t &box = feature.box;

    //const gpu_integral_channels_t::Type  // could cause overflows during a + c
    const float
            a = tex2D(tex, x + box.min_corner().x(), box.min_corner().y() + y_offset), // top left
            b = tex2D(tex, x + box.max_corner().x(), box.min_corner().y() + y_offset), // top right
            c = tex2D(tex, x + box.max_corner().x(), box.max_corner().y() + y_offset), // bottom right
            d = tex2D(tex, x + box.min_corner().x(), box.max_corner().y() + y_offset); // bottom left
#undef tex

    const float feature_value = a +c -b -d;

    return feature_value;
}


template <typename FeatureType>
inline
__device__
float get_feature_value_tex2d(const FeatureType &feature,
                              const int x, const int y,
                              const gpu_2d_integral_channels_t::KernelConstData &integral_channels)
{
    // if x or y are too high, some of these indices may be fall outside the channel memory

    //const size_t integral_channels_height = integral_channels.size[1];
    const size_t integral_channels_height = 120; // FIXME HARDCODED TEST (will only work with shrinking factor 4 and images of size 640x480
    const float y_offset = y + feature.channel_index*integral_channels_height;

    // in CUDA 5 (4.2 ?) references to textures are not allowed, we use macro work around
    //    gpu_integral_channels_2d_texture_t &tex = integral_channels_2d_texture;
#define tex integral_channels_2d_texture

    const typename FeatureType::rectangle_t &box = feature.box;

    //const gpu_integral_channels_t::Type  // could cause overflows during a + c
    const float
            a = tex2D(tex, x + box.min_corner().x(), box.min_corner().y() + y_offset), // top left
            b = tex2D(tex, x + box.max_corner().x(), box.min_corner().y() + y_offset), // top right
            c = tex2D(tex, x + box.max_corner().x(), box.max_corner().y() + y_offset), // bottom right
            d = tex2D(tex, x + box.min_corner().x(), box.max_corner().y() + y_offset); // bottom left
#undef tex

    const float feature_value = a +c -b -d;

    return feature_value;
}


template <typename FeatureType, bool should_use_2d_texture>
inline
__device__
float get_feature_value(const FeatureType &feature,
                        const int x, const int y,
                        const gpu_integral_channels_t::KernelConstData &integral_channels)
{
    // Global memory is the "naive way" of coding,
    // in GPU doing (1d or 2d) texture mapping will give you speed gains with respect to global
    return get_feature_value_global_memory(feature, x, y, integral_channels);

    // default implementation (hopefully optimized by the compiler)
    if (should_use_2d_texture)
    {
        return get_feature_value_tex2d(feature, x, y, integral_channels);
    }
    else
    {
        //return get_feature_value_global_memory(feature, x, y, integral_channels);
        return get_feature_value_tex1d(feature, x, y, integral_channels);
    }

    //return 0;
}


// >>> evaluate_decision_stump over SimpleDecisionStump and DecisionStumpWithWeights are the ones used in the excercise <<<

inline
__device__
bool evaluate_decision_stump(const SimpleDecisionStump &stump,
                             const float &feature_value)
{
    // uses >= to be consistent with Markus Mathias code
    return (feature_value >= stump.feature_threshold);
}


inline
__device__
float evaluate_decision_stump(const DecisionStumpWithWeights &stump,
                              const float &feature_value)
{
    // uses >= to be consistent with Markus Mathias code
    return (feature_value >= stump.feature_threshold)? stump.weight_true_leaf : stump.weight_false_leaf;
}





inline
__device__
void update_detection_score_v0(
        const int x, const int y,
        const cascade_stage_t &stage,
        const gpu_integral_channels_t::KernelConstData &integral_channels,
        float &current_score)
{
    const cascade_stage_t::weak_classifier_t &weak_classifier = stage.weak_classifier;
    typedef cascade_stage_t::weak_classifier_t::feature_t feature_t;

    // level 1 nodes evaluation returns a boolean value,
    // level 2 nodes evaluation returns directly the float value to add to the score

    const float level1_feature_value =
            get_feature_value<feature_t, use_2d_texture>(
                weak_classifier.level1_node.feature, x, y, integral_channels);

    if(evaluate_decision_stump(weak_classifier.level1_node, level1_feature_value))
    {
        const float level2_true_feature_value =
                get_feature_value<feature_t, use_2d_texture>(
                    weak_classifier.level2_true_node.feature, x, y, integral_channels);

        current_score += evaluate_decision_stump(weak_classifier.level2_true_node, level2_true_feature_value);
    }
    else
    {
        const float level2_false_feature_value =
                get_feature_value<feature_t, use_2d_texture>(
                    weak_classifier.level2_false_node.feature, x, y, integral_channels);

        current_score += evaluate_decision_stump(weak_classifier.level2_false_node, level2_false_feature_value);
    }

    return;
}



/// type int because atomicAdd does not support size_t
__device__ int num_gpu_detections[1];
Cuda::Symbol<int, 1> num_gpu_detections_symbol(Cuda::Size<1>(1), num_gpu_detections);
int num_detections_int;
Cuda::HostMemoryReference1D<int> num_detections_host_ref(1, &num_detections_int);


void move_num_detections_from_cpu_to_gpu(size_t &num_detections)
{ // move num_detections from CPU to GPU --
    num_detections_int = static_cast<int>(num_detections);
    Cuda::copy(num_gpu_detections_symbol, num_detections_host_ref);
    return;
}

void move_num_detections_from_gpu_to_cpu(size_t &num_detections)
{ // move (updated) num_detections from GPU to CPU
    Cuda::copy(num_detections_host_ref, num_gpu_detections_symbol);
    if(num_detections_int < static_cast<int>(num_detections))
    {
        throw std::runtime_error("Something went terribly wrong when updating the number of gpu detections");
    }
    num_detections = static_cast<size_t>(num_detections_int);
    return;
}



template<typename ScaleType>
inline
__device__
void add_detection(
        gpu_detections_t::KernelData &gpu_detections,
        const int x, const int y, const ScaleType scale_index,
        const float detection_score)
{
    gpu_detection_t detection;
    detection.scale_index = static_cast<boost::int16_t>(scale_index);
    detection.x = static_cast<boost::int16_t>(x);
    detection.y = static_cast<boost::int16_t>(y);
    detection.score = detection_score;

    const size_t detection_index = atomicAdd(num_gpu_detections, 1);
    if(detection_index < gpu_detections.size[0])
    {
        // copy the detection into the global memory
        gpu_detections.data[detection_index] = detection;
    }
    else
    {
        // we drop out of range detections
    }

    return;
}





/// helper method to map the device memory to the specific texture reference
/// This specific implementation will do a 1d binding
void bind_integral_channels_to_1d_texture(gpu_integral_channels_t &integral_channels)
{

    //integral_channels_texture.filterMode = cudaFilterModeLinear; // linear interpolation of the values
    integral_channels_1d_texture.filterMode = cudaFilterModePoint; // normal access to the values
    //integral_channels.bindTexture(integral_channels_texture);

    // cuda does not support binding 3d memory data.
    // We will hack this and bind the 3d data, as if it was 1d data,
    // and then have ad-hoc texture access in the kernel
    // (if interpolation is needed, will need to do a different 2d data hack
    const cudaChannelFormatDesc texture_channel_description = \
            cudaCreateChannelDesc<gpu_integral_channels_t::Type>();

    if(texture_channel_description.f == cudaChannelFormatKindNone
            or texture_channel_description.f != cudaChannelFormatKindUnsigned )
    {
        throw std::runtime_error("cudaCreateChannelDesc failed");
    }

    if(false)
    {
        printf("texture_channel_description.x == %i\n", texture_channel_description.x);
        printf("texture_channel_description.y == %i\n", texture_channel_description.y);
        printf("texture_channel_description.z == %i\n", texture_channel_description.z);
        printf("texture_channel_description.w == %i\n", texture_channel_description.w);
    }

    // FIXME add this 3D to 1D strategy into cudatemplates
    CUDA_CHECK(cudaBindTexture(0, integral_channels_1d_texture, integral_channels.getBuffer(),
                               texture_channel_description, integral_channels.getBytes()));

    cuda_safe_call( cudaGetLastError() );

    return;
}


/// helper method to map the device memory to the specific texture reference
/// This specific implementation will do a 2d binding
void bind_integral_channels_to_2d_texture(gpu_3d_integral_channels_t &integral_channels)
{ // 3d integral channels case

    // linear interpolation of the values, only valid for floating point types
    //integral_channels_2d_texture.filterMode = cudaFilterModeLinear;
    integral_channels_2d_texture.filterMode = cudaFilterModePoint; // normal access to the values
    //integral_channels.bindTexture(integral_channels_2d_texture);

    // cuda does not support binding 3d memory data.
    // We will hack this and bind the 3d data, as if it was 2d data,
    // and then have ad-hoc texture access in the kernel
    const cudaChannelFormatDesc texture_channel_description = cudaCreateChannelDesc<gpu_3d_integral_channels_t::Type>();

    if(texture_channel_description.f == cudaChannelFormatKindNone
            or texture_channel_description.f != cudaChannelFormatKindUnsigned )
    {
        throw std::runtime_error("cudaCreateChannelDesc seems to have failed");
    }

    if(false)
    {
        printf("texture_channel_description.x == %i\n", texture_channel_description.x);
        printf("texture_channel_description.y == %i\n", texture_channel_description.y);
        printf("texture_channel_description.z == %i\n", texture_channel_description.z);
        printf("texture_channel_description.w == %i\n", texture_channel_description.w);
    }

    // Layout.size is width, height, num_channels
    const size_t
            integral_channels_width = integral_channels.getLayout().size[0],
            integral_channels_height = integral_channels.getLayout().size[1],
            num_integral_channels = integral_channels.getLayout().size[2],
            //channel_stride = integral_channels.getLayout().stride[1],
            row_stride = integral_channels.getLayout().stride[0],
            pitch_in_bytes = row_stride * sizeof(gpu_3d_integral_channels_t::Type);

    if(false)
    {
        printf("image width/height == %zi, %zi; row_stride == %zi\n",
               integral_channels.getLayout().size[0], integral_channels.getLayout().size[1],
                integral_channels.getLayout().stride[0]);

        printf("integral_channels size / channel_stride == %.3f\n",
               integral_channels.getLayout().stride[2] / float(integral_channels.getLayout().stride[1]) );
    }

    // FIXME add this 3D to 2D strategy into cudatemplates
    CUDA_CHECK(cudaBindTexture2D(0, integral_channels_2d_texture, integral_channels.getBuffer(),
                                 texture_channel_description,
                                 integral_channels_width, integral_channels_height*num_integral_channels,
                                 pitch_in_bytes));

    cuda_safe_call( cudaGetLastError() );
    return;
}


/// helper method to map the device memory to the specific texture reference
/// This specific implementation will do a 2d binding
void bind_integral_channels_to_2d_texture(gpu_2d_integral_channels_t &integral_channels)
{ // 2d integral channels case

    // linear interpolation of the values, only valid for floating point types
    //integral_channels_2d_texture.filterMode = cudaFilterModeLinear;
    integral_channels_2d_texture.filterMode = cudaFilterModePoint; // normal access to the values

    // integral_channels_height == (channel_height*num_channels) + 1
    // 2d to 2d binding
    integral_channels.bindTexture(integral_channels_2d_texture);

    cuda_safe_call( cudaGetLastError() );
    return;
}



void bind_integral_channels_texture(gpu_integral_channels_t &integral_channels)
{
    if(use_2d_texture)
    {
        bind_integral_channels_to_2d_texture(integral_channels);
    }
    else
    {
        bind_integral_channels_to_1d_texture(integral_channels);
    }

    return;
}


void unbind_integral_channels_texture()
{

    if(use_2d_texture)
    {
        cuda_safe_call( cudaUnbindTexture(integral_channels_2d_texture) );
    }
    else
    {
        cuda_safe_call( cudaUnbindTexture(integral_channels_1d_texture) );
    }
    cuda_safe_call( cudaGetLastError() );

    return;
}




// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


/// This kernel is called for each position where we which to detect objects
/// we assume that the border effects where already checked when computing the DetectorSearchRange
/// thus we do not do any checks here.
/// This kernel is based on integral_channels_detector_kernel,
/// but does detections for all scales instead of a single scale
/// @see IntegralChannelsDetector
__global__
void integral_channels_detector_over_all_scales_kernel_v0(
        const gpu_integral_channels_t::KernelConstData integral_channels,
        const int max_search_range_min_x, const int max_search_range_min_y,
        const int /*max_search_range_width*/, const int /*max_search_range_height*/,
        const gpu_scales_data_t::KernelConstData scales_data,
        const typename Cuda::DeviceMemory<cascade_stage_t, 2>::KernelConstData detection_cascade_per_scale,
        const float score_threshold,
        gpu_detections_t::KernelData gpu_detections)
{
    const int
            delta_x = blockIdx.x * blockDim.x + threadIdx.x,
            delta_y = blockIdx.y * blockDim.y + threadIdx.y,
            x = max_search_range_min_x + delta_x,
            y = max_search_range_min_y + delta_y;

    float max_detection_score = -1E10; // initialized with a very negative value
    int max_detection_scale_index = 0;

    // (we use int instead of size_t, as indicated by Cuda Best Programming Practices, section 6.3)
    // (using int of size_t moved from 1.48 Hz to 1.53 Hz)

    const int
            cascade_length = detection_cascade_per_scale.size[0],
            num_scales = scales_data.size[0];

    for(int scale_index=0; scale_index < num_scales; scale_index +=1)
    {
        // when using softcascades, __syncthreads here is a significant slow down, so not using it

        // we copy the search stage from global memory to thread memory
        // (here using copy or reference seems to make no difference, 11.82 Hz in both cases)
        const gpu_scale_datum_t::search_range_t search_range = scales_data.data[scale_index].search_range;
        // (in current code, we ignore the gpu_scale_datum_t stride value)

        // we order the if conditions putting most likelly ones first
        if( (y > search_range.max_corner().y())
                or (y < search_range.min_corner().y())
                or (x < search_range.min_corner().x())
                or (x > search_range.max_corner().x()) )
        {
            // current pixel is out of this scale search range, we skip computations
            // (nothing to do here)
        }
        else
        { // inside search range

            // retrieve current score value
            float detection_score = 0;

            const int scale_offset = scale_index * detection_cascade_per_scale.stride[0];

            for(int stage_index=0; stage_index < cascade_length; stage_index+=1)
            {
                const int index = scale_offset + stage_index;

                // we copy the cascade stage from global memory to thread memory
                // (when using a reference code runs at ~4.35 Hz, with copy it runs at ~4.55 Hz)
                const cascade_stage_t stage = detection_cascade_per_scale.data[index];

                update_detection_score_v0(x, y, stage, integral_channels, detection_score);

            } // end of "for each stage"

            if(detection_score > max_detection_score)
            {
                max_detection_score = detection_score;
                max_detection_scale_index = scale_index;
            }

        } // end of "inside search range or not"

    } // end of "for each scale"


    // >= to be consistent with Markus's code
    if(max_detection_score >= score_threshold)
    {
        // we got a detection
        add_detection(gpu_detections, x, y, max_detection_scale_index, max_detection_score);

    } // end of "if detection score is high enough"

    return;
} // end of integral_channels_detector_over_all_scales_kernel_v0




/// this method directly adds elements into the gpu_detections vector
void integral_channels_detector_over_all_scales_impl_v0(
        gpu_integral_channels_t &integral_channels,
        gpu_scale_datum_t::search_range_t &max_search_range,
        gpu_scales_data_t &scales_data,
        gpu_detection_cascade_per_scale_t &detection_cascade_per_scale,
        const float score_threshold,
        const bool use_the_model_cascade,
        gpu_detections_t& gpu_detections,
        size_t &num_detections)
{

    const int
            max_search_range_min_x = max_search_range.min_corner().x(),
            max_search_range_min_y = max_search_range.min_corner().y(),
            max_search_range_width = max_search_range.max_corner().x() - max_search_range_min_x,
            max_search_range_height = max_search_range.max_corner().y() - max_search_range_min_y;

    // >>> you should use the CUDA occupancy calculator to define the number of threads and block size <<<
    const int
            num_threads = 12,
            block_y = 4, block_x = num_threads / block_y;

    dim3 block_dimensions(block_x, block_y);

    dim3 grid_dimensions(div_up(max_search_range_width, block_dimensions.x),
                         div_up(max_search_range_height, block_dimensions.y));

    // prepare variables for kernel call --
    bind_integral_channels_texture(integral_channels);
    move_num_detections_from_cpu_to_gpu(num_detections);

    // call the GPU kernel --

    if(use_the_model_cascade)
    {
        throw std::runtime_error("You must implement the soft cascade usage");
        // if detection_score <= stage.cascade_threshold then "this candidate detection window can be dropped"

    }
    else
    {
        integral_channels_detector_over_all_scales_kernel_v0
                <<<grid_dimensions, block_dimensions>>> (
                                                          integral_channels,
                                                          max_search_range_min_x, max_search_range_min_y,
                                                          max_search_range_width, max_search_range_height,
                                                          scales_data,
                                                          detection_cascade_per_scale,
                                                          score_threshold,
                                                          gpu_detections);
    }

    cuda_safe_call( cudaGetLastError() );
    cuda_safe_call( cudaDeviceSynchronize() );

    // clean-up variables after kernel call --
    unbind_integral_channels_texture();
    move_num_detections_from_gpu_to_cpu(num_detections);

    return;
}




void integral_channels_detector_over_all_scales(
        gpu_integral_channels_t &integral_channels,
        gpu_scale_datum_t::search_range_t &max_search_range,
        const int max_search_range_width, const int max_search_range_height,
        gpu_scales_data_t &scales_data,
        gpu_detection_cascade_per_scale_t &detection_cascade_per_scale,
        const float score_threshold,
        const bool use_the_model_cascade,
        gpu_detections_t& gpu_detections,
        size_t &num_detections)
{
    // v0 runs at about ~0.2 Hz,
    // the fastest code that uses exactly the same inputs, runs at about 40~50 Hz, can you match it ?
    // Write your own integral_channels_detector_over_all_scales_impl_v1, v2, vN !
    integral_channels_detector_over_all_scales_impl_v0(
                integral_channels,
                max_search_range, scales_data,
                detection_cascade_per_scale,
                score_threshold, use_the_model_cascade, gpu_detections, num_detections);

    return;
}

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// >>> Everything below this line is irrelevant for the excercise <<<


#include "./integral_channels_detector_not_relevant_for_exercise.cuda_include"



} // end of namespace objects_detection
} // end of namespace doppia


